/* DebugEventsPlugin
 *
 * Simple NPAPI plugin for testing the browser/plugin interface, especially
 * to see what kinds of events the browser passes to the plugin.
 *
 * The original template was the NetscapeCocoaPlugin sample that comes with
 * the WebKit source distro (in WebKitExamplePlugins).  QuickDraw support
 * was scavanged from various places in the Mozilla tree -- particularly the
 * Null/Default plugin.
 */

#define XP_MACOSX 1

#import "npapi.h"
#import "npfunctions.h"
#import "npruntime.h"
#import "npupp.h"

#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>

#define EXPORT __attribute__((visibility("default")))

#define SUPPORT_CORE_GRAHPICS "supportCoreGraphics"
#define SUPPORT_COCOA "supportCocoa"

// Uncomment the following line to log mouse-moved events
//#define LOG_MOUSE_MOVED_EVENTS 1

// Class for a static "handler" object that detects when the plugin is loaded
// or unloaded.
class loadHandler
{
public:
  loadHandler();
  ~loadHandler();
};

loadHandler::loadHandler()
{
    NSAutoreleasePool *localPool = [[NSAutoreleasePool alloc] init];
    if (localPool) {
        NSLog(@"DebugEventsPlugin loaded");
        [localPool release];
    }
}

loadHandler::~loadHandler()
{
    NSAutoreleasePool *localPool = [[NSAutoreleasePool alloc] init];
    if (localPool) {
        NSLog(@"DebugEventsPlugin unloaded");
        [localPool release];
    }
}

loadHandler handler = loadHandler();

@interface NSWindow (DebugEventsPlugin_Undocumented)
- (BOOL)_hasWindowRef;
@end

@interface NSWindow (DebugEventsPlugin)
+ (WindowRef) existingCarbonWindowFor:(NSWindow *)cocoaWindow;
@end

@implementation NSWindow (DebugEventsPlugin)

+ (WindowRef) existingCarbonWindowFor:(NSWindow *)cocoaWindow
{
    if (![cocoaWindow _hasWindowRef])
        return nil;
    return (WindowRef) [cocoaWindow windowRef];
}

@end

#ifdef __ppc__

// glue for mapping outgoing Macho function pointers to TVectors
struct TFPtoTVGlue{
    void* glue[2];
};

struct pluginFuncsGlueTable {
    TFPtoTVGlue     newp;
    TFPtoTVGlue     destroy;
    TFPtoTVGlue     setwindow;
    TFPtoTVGlue     newstream;
    TFPtoTVGlue     destroystream;
    TFPtoTVGlue     asfile;
    TFPtoTVGlue     writeready;
    TFPtoTVGlue     write;
    TFPtoTVGlue     print;
    TFPtoTVGlue     event;
    TFPtoTVGlue     urlnotify;
    TFPtoTVGlue     javaClass;
    TFPtoTVGlue     getvalue;
    TFPtoTVGlue     setvalue;
    TFPtoTVGlue     shutdown;
} gPluginFuncsGlueTable;

static inline void* SetupFPtoTVGlue(TFPtoTVGlue* functionGlue, void* fp)
{
    functionGlue->glue[0] = fp;
    functionGlue->glue[1] = 0;
    return functionGlue;
}

#define PLUGIN_TO_HOST_GLUE(name, fp) (SetupFPtoTVGlue(&gPluginFuncsGlueTable.name, (void*)fp))

// glue for mapping netscape TVectors to Macho function pointers
struct TTVtoFPGlue {
    uint32_t glue[6];
};

struct netscapeFuncsGlueTable {
    TTVtoFPGlue             geturl;
    TTVtoFPGlue             posturl;
    TTVtoFPGlue             requestread;
    TTVtoFPGlue             newstream;
    TTVtoFPGlue             write;
    TTVtoFPGlue             destroystream;
    TTVtoFPGlue             status;
    TTVtoFPGlue             uagent;
    TTVtoFPGlue             memalloc;
    TTVtoFPGlue             memfree;
    TTVtoFPGlue             memflush;
    TTVtoFPGlue             reloadplugins;
    TTVtoFPGlue             getJavaEnv;
    TTVtoFPGlue             getJavaPeer;
    TTVtoFPGlue             geturlnotify;
    TTVtoFPGlue             posturlnotify;
    TTVtoFPGlue             getvalue;
    TTVtoFPGlue             setvalue;
    TTVtoFPGlue             invalidaterect;
    TTVtoFPGlue             invalidateregion;
    TTVtoFPGlue             forceredraw;
    TTVtoFPGlue             getstringidentifier;
    TTVtoFPGlue             getstringidentifiers;
    TTVtoFPGlue             getintidentifier;
    TTVtoFPGlue             identifierisstring;
    TTVtoFPGlue             utf8fromidentifier;
    TTVtoFPGlue             intfromidentifier;
    TTVtoFPGlue             createobject;
    TTVtoFPGlue             retainobject;
    TTVtoFPGlue             releaseobject;
    TTVtoFPGlue             invoke;
    TTVtoFPGlue             invokeDefault;
    TTVtoFPGlue             evaluate;
    TTVtoFPGlue             getproperty;
    TTVtoFPGlue             setproperty;
    TTVtoFPGlue             removeproperty;
    TTVtoFPGlue             hasproperty;
    TTVtoFPGlue             hasmethod;
    TTVtoFPGlue             releasevariantvalue;
    TTVtoFPGlue             setexception;
    TTVtoFPGlue             pushpopupsenabledstate;
    TTVtoFPGlue             poppopupsenabledstate;
    TTVtoFPGlue             enumerate;
    TTVtoFPGlue             pluginthreadasynccall;
    TTVtoFPGlue             construct;
    TTVtoFPGlue             getvalueforurl;
    TTVtoFPGlue             setvalueforurl;
    TTVtoFPGlue             getauthenticationinfo;
    TTVtoFPGlue             scheduletimer;
    TTVtoFPGlue             unscheduletimer;
    TTVtoFPGlue             popupcontextmenu;
    TTVtoFPGlue             convertpoint;
} gNetscapeFuncsGlueTable;

static void* SetupTVtoFPGlue(TTVtoFPGlue* functionGlue, void* tvp)
{
    static const TTVtoFPGlue glueTemplate = { 0x3D800000, 0x618C0000, 0x800C0000, 0x804C0004, 0x7C0903A6, 0x4E800420 };

    memcpy(functionGlue, &glueTemplate, sizeof(TTVtoFPGlue));
    functionGlue->glue[0] |= ((UInt32)tvp >> 16);
    functionGlue->glue[1] |= ((UInt32)tvp & 0xFFFF);
    ::MakeDataExecutable(functionGlue, sizeof(TTVtoFPGlue));
    return functionGlue;
}

#define HOST_TO_PLUGIN_GLUE(name, fp) (SetupTVtoFPGlue(&gNetscapeFuncsGlueTable.name, (void*)fp))

#else

#define PLUGIN_TO_HOST_GLUE(name, fp) (fp)
#define HOST_TO_PLUGIN_GLUE(name, fp) (fp)

#endif /* __ppc__ */

// Browser function table
static NPNetscapeFuncs browser = {0};

// Support for scriptable npruntime object

typedef struct TestNPObject : NPObject {
  NPP npp;
} TestNPObject;

static NPClass sSOClass = {0};

NPObject* scriptableAllocate(NPP npp, NPClass* aClass);
void scriptableDeallocate(NPObject* npobj);
void scriptableInvalidate(NPObject* npobj);
bool scriptableHasMethod(NPObject* npobj, NPIdentifier name);
bool scriptableInvoke(NPObject* npobj, NPIdentifier name, const NPVariant* args, uint32_t argCount, NPVariant* result);
bool scriptableInvokeDefault(NPObject* npobj, const NPVariant* args, uint32_t argCount, NPVariant* result);
bool scriptableHasProperty(NPObject* npobj, NPIdentifier name);
bool scriptableGetProperty(NPObject* npobj, NPIdentifier name, NPVariant* result);
bool scriptableSetProperty(NPObject* npobj, NPIdentifier name, const NPVariant* value);
bool scriptableRemoveProperty(NPObject* npobj, NPIdentifier name);
bool scriptableEnumerate(NPObject* npobj, NPIdentifier** identifier, uint32_t* count);
bool scriptableConstruct(NPObject* npobj, const NPVariant* args, uint32_t argCount, NPVariant* result);

// Structure for per-instance storage
typedef struct PluginObject
{
    NPP npp;

    NPWindow window;
    uint16_t mode;

    NSString *debugInfo;

    NPBool supportsCoreGraphics;
    NPBool supportsCocoa;

    NPBool hasFocus;
    NPBool windowHasFocus;

    TestNPObject *scriptableObject;

    GrafPtr savedPortQD;
    RgnHandle savedClipQD;
    short savedPortTopQD;
    short savedPortLeftQD;
} PluginObject;

NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16_t mode, int16_t argc, char* argn[], char* argv[], NPSavedData* saved);
NPError NPP_Destroy(NPP instance, NPSavedData** save);
NPError NPP_SetWindow(NPP instance, NPWindow* window);
NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16_t* stype);
NPError NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason);
int32_t NPP_WriteReady(NPP instance, NPStream* stream);
int32_t NPP_Write(NPP instance, NPStream* stream, int32_t offset, int32_t len, void* buffer);
void NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname);
void NPP_Print(NPP instance, NPPrint* platformPrint);
int16_t NPP_HandleEvent(NPP instance, void* event);
void NPP_URLNotify(NPP instance, const char* URL, NPReason reason, void* notifyData);
NPError NPP_GetValue(NPP instance, NPPVariable variable, void *value);
NPError NPP_SetValue(NPP instance, NPNVariable variable, void *value);

// Mach-o entry points
extern "C" {
EXPORT int main(NPNetscapeFuncs* browserFuncs, NPPluginFuncs* pluginFuncs, NPP_ShutdownUPP* unloadUpp);
EXPORT NPError NP_Initialize(NPNetscapeFuncs *browserFuncs);
EXPORT NPError NP_GetEntryPoints(NPPluginFuncs *pluginFuncs);
EXPORT void NP_Shutdown(void);
}

// Needed for old fashioned browsers ... like FF2 and Opera.
int main(NPNetscapeFuncs* browserFuncs, NPPluginFuncs* pluginFuncs, NPP_ShutdownUPP* unloadUpp)
{
    NSLog(@"DebugEventsPlugin: main(): browserFuncs 0x%.8X, pluginFuncs 0x%.8X, unloadUpp 0x%.8X",
          (unsigned) browserFuncs, (unsigned) pluginFuncs, (unsigned) unloadUpp);
    if (!browserFuncs || !pluginFuncs || !unloadUpp) {
        NSLog(@"DebugEventsPlugin: main() error return");
        return NPERR_INVALID_FUNCTABLE_ERROR;
    }

    browser.version             = browserFuncs->version;
    browser.size                = browserFuncs->size;
    browser.memalloc            = (NPN_MemAllocUPP)HOST_TO_PLUGIN_GLUE(memalloc, browserFuncs->memalloc);
    browser.memfree             = (NPN_MemFreeUPP)HOST_TO_PLUGIN_GLUE(memfree, browserFuncs->memfree);
    browser.getvalue            = (NPN_GetValueUPP)HOST_TO_PLUGIN_GLUE(getvalue, browserFuncs->getvalue);
    browser.setvalue            = (NPN_SetValueUPP)HOST_TO_PLUGIN_GLUE(setvalue, browserFuncs->setvalue);
    browser.invalidaterect      = (NPN_InvalidateRectUPP)HOST_TO_PLUGIN_GLUE(invalidaterect, browserFuncs->invalidaterect);
    browser.getstringidentifier = (NPN_GetStringIdentifierUPP)HOST_TO_PLUGIN_GLUE(getstringidentifier, browserFuncs->getstringidentifier);
    browser.createobject        = (NPN_CreateObjectUPP)HOST_TO_PLUGIN_GLUE(createobject, browserFuncs->createobject);
    browser.retainobject        = (NPN_RetainObjectUPP)HOST_TO_PLUGIN_GLUE(retainobject, browserFuncs->retainobject);
    browser.releaseobject       = (NPN_ReleaseObjectUPP)HOST_TO_PLUGIN_GLUE(releaseobject, browserFuncs->releaseobject);
    browser.invoke              = (NPN_InvokeUPP)HOST_TO_PLUGIN_GLUE(invoke, browserFuncs->invoke);
    browser.getproperty         = (NPN_GetPropertyUPP)HOST_TO_PLUGIN_GLUE(getproperty, browserFuncs->getproperty);
    browser.releasevariantvalue = (NPN_ReleaseVariantValueUPP)HOST_TO_PLUGIN_GLUE(releasevariantvalue, browserFuncs->releasevariantvalue);

    pluginFuncs->version        = 14;
    pluginFuncs->size           = sizeof(pluginFuncs);
    pluginFuncs->newp           = NewNPP_NewProc(PLUGIN_TO_HOST_GLUE(newp, NPP_New));
    pluginFuncs->destroy        = NewNPP_DestroyProc(PLUGIN_TO_HOST_GLUE(destroy, NPP_Destroy));
    pluginFuncs->setwindow      = NewNPP_SetWindowProc(PLUGIN_TO_HOST_GLUE(setwindow, NPP_SetWindow));
    pluginFuncs->newstream      = NewNPP_NewStreamProc(PLUGIN_TO_HOST_GLUE(newstream, NPP_NewStream));
    pluginFuncs->destroystream  = NewNPP_DestroyStreamProc(PLUGIN_TO_HOST_GLUE(destroystream, NPP_DestroyStream));
    pluginFuncs->asfile         = NewNPP_StreamAsFileProc(PLUGIN_TO_HOST_GLUE(asfile, NPP_StreamAsFile));
    pluginFuncs->writeready     = NewNPP_WriteReadyProc(PLUGIN_TO_HOST_GLUE(writeready, NPP_WriteReady));
    pluginFuncs->write          = (NPP_WriteProcPtr)NewNPP_WriteProc(PLUGIN_TO_HOST_GLUE(write, NPP_Write));
    pluginFuncs->print          = NewNPP_PrintProc(PLUGIN_TO_HOST_GLUE(print, NPP_Print));
    pluginFuncs->event          = NewNPP_HandleEventProc(PLUGIN_TO_HOST_GLUE(event, NPP_HandleEvent));
    pluginFuncs->urlnotify      = NewNPP_URLNotifyProc(PLUGIN_TO_HOST_GLUE(urlnotify, NPP_URLNotify));
    pluginFuncs->getvalue       = NewNPP_GetValueProc(PLUGIN_TO_HOST_GLUE(getvalue, NPP_GetValue));
    pluginFuncs->setvalue       = NewNPP_SetValueProc(PLUGIN_TO_HOST_GLUE(setvalue, NPP_SetValue));

    *unloadUpp                  = NewNPP_ShutdownProc(PLUGIN_TO_HOST_GLUE(shutdown, NP_Shutdown));

    sSOClass.structVersion      = NP_CLASS_STRUCT_VERSION;
    sSOClass.allocate           = (NPAllocateFunctionPtr)scriptableAllocate;
    sSOClass.deallocate         = (NPDeallocateFunctionPtr)scriptableDeallocate;
    sSOClass.invalidate         = (NPInvalidateFunctionPtr)scriptableInvalidate;
    sSOClass.hasMethod          = (NPHasMethodFunctionPtr)scriptableHasMethod;
    sSOClass.invoke             = (NPInvokeFunctionPtr)scriptableInvoke;
    sSOClass.invokeDefault      = (NPInvokeDefaultFunctionPtr)scriptableInvokeDefault;
    sSOClass.hasProperty        = (NPHasPropertyFunctionPtr)scriptableHasProperty;
    sSOClass.getProperty        = (NPGetPropertyFunctionPtr)scriptableGetProperty;
    sSOClass.setProperty        = (NPSetPropertyFunctionPtr)scriptableSetProperty;
    sSOClass.removeProperty     = (NPRemovePropertyFunctionPtr)scriptableRemoveProperty;
    sSOClass.enumerate          = (NPEnumerationFunctionPtr)scriptableEnumerate;
    sSOClass.construct          = (NPConstructFunctionPtr)scriptableConstruct;

    return NPERR_NO_ERROR;
}

NPError NP_Initialize(NPNetscapeFuncs* browserFuncs)
{
    NSLog(@"DebugEventsPlugin: NP_Initialize(): browserFuncs 0x%.8X", (unsigned) browserFuncs);
    if (!browserFuncs) {
        NSLog(@"DebugEventsPlugin: NP_Initialize() error return");
        return NPERR_INVALID_FUNCTABLE_ERROR;
    }

    browser.version = browserFuncs->version;
    browser.size = browserFuncs->size;
    browser.memalloc = browserFuncs->memalloc;
    browser.memfree = browserFuncs->memfree;
    browser.getvalue = browserFuncs->getvalue;
    browser.setvalue = browserFuncs->setvalue;
    browser.invalidaterect = browserFuncs->invalidaterect;
    browser.getstringidentifier = browserFuncs->getstringidentifier;
    browser.createobject = browserFuncs->createobject;
    browser.retainobject = browserFuncs->retainobject;
    browser.releaseobject = browserFuncs->releaseobject;
    browser.invoke = browserFuncs->invoke;
    browser.getproperty = browserFuncs->getproperty;
    browser.releasevariantvalue = browserFuncs->releasevariantvalue;

    sSOClass.structVersion      = NP_CLASS_STRUCT_VERSION;
    sSOClass.allocate           = (NPAllocateFunctionPtr)scriptableAllocate;
    sSOClass.deallocate         = (NPDeallocateFunctionPtr)scriptableDeallocate;
    sSOClass.invalidate         = (NPInvalidateFunctionPtr)scriptableInvalidate;
    sSOClass.hasMethod          = (NPHasMethodFunctionPtr)scriptableHasMethod;
    sSOClass.invoke             = (NPInvokeFunctionPtr)scriptableInvoke;
    sSOClass.invokeDefault      = (NPInvokeDefaultFunctionPtr)scriptableInvokeDefault;
    sSOClass.hasProperty        = (NPHasPropertyFunctionPtr)scriptableHasProperty;
    sSOClass.getProperty        = (NPGetPropertyFunctionPtr)scriptableGetProperty;
    sSOClass.setProperty        = (NPSetPropertyFunctionPtr)scriptableSetProperty;
    sSOClass.removeProperty     = (NPRemovePropertyFunctionPtr)scriptableRemoveProperty;
    sSOClass.enumerate          = (NPEnumerationFunctionPtr)scriptableEnumerate;
    sSOClass.construct          = (NPConstructFunctionPtr)scriptableConstruct;

    return NPERR_NO_ERROR;
}

NPError NP_GetEntryPoints(NPPluginFuncs* pluginFuncs)
{
    NSLog(@"DebugEventsPlugin: NP_GetEntryPoints(): pluginFuncs 0x%.8X", (unsigned) pluginFuncs);
    if (!pluginFuncs) {
        NSLog(@"DebugEventsPlugin: NP_GetEntryPoints() error return");
        return NPERR_INVALID_FUNCTABLE_ERROR;
    }

    pluginFuncs->version = 14;
    pluginFuncs->size = sizeof(pluginFuncs);
    pluginFuncs->newp = NPP_New;
    pluginFuncs->destroy = NPP_Destroy;
    pluginFuncs->setwindow = NPP_SetWindow;
    pluginFuncs->newstream = NPP_NewStream;
    pluginFuncs->destroystream = NPP_DestroyStream;
    pluginFuncs->asfile = NPP_StreamAsFile;
    pluginFuncs->writeready = NPP_WriteReady;
    pluginFuncs->write = (NPP_WriteProcPtr)NPP_Write;
    pluginFuncs->print = NPP_Print;
    pluginFuncs->event = NPP_HandleEvent;
    pluginFuncs->urlnotify = NPP_URLNotify;
    pluginFuncs->getvalue = NPP_GetValue;
    pluginFuncs->setvalue = NPP_SetValue;

    return NPERR_NO_ERROR;
}

void NP_Shutdown(void)
{
    NSLog(@"DebugEventsPlugin: NP_Shutdown()");
}

static void checkForOverride(char *setting, NPBool *value, NPBool *valueFound,
                             int16_t argc, char* argn[], char* argv[])
{
    if (!valueFound)
        return;
    *valueFound = FALSE;
    if (!setting || !value || !argc || !argn || !argv)
        return;

    for (int16_t i = 0; i < argc; ++i) {
        if (strcasecmp(setting, argn[i]))
            continue;
        if (!strcasecmp(argv[i], "0") || !strcasecmp(argv[i], "false") || !strcasecmp(argv[i], "no")) {
            *valueFound = TRUE;
            *value = FALSE;
            return;
        }
        if (!strcasecmp(argv[i], "1") || !strcasecmp(argv[i], "true") || !strcasecmp(argv[i], "yes")) {
            *valueFound = TRUE;
            *value = TRUE;
            return;
        }
    }
}

static char *modeString(uint16_t mode)
{
    switch (mode) {
        case NP_EMBED: return "NP_EMBED";
        case NP_FULL: return "NP_FULL";
        default: return "Unknown";
    }
}

static char *typeString(NPWindowType type)
{
    switch (type) {
        case NPWindowTypeWindow: return "NPWindowTypeWindow";
        case NPWindowTypeDrawable: return "NPWindowTypeDrawable";
        default: return "Unknown";
    }
}

static char *getUTF8String(NPVariant stringObj)
{
    if (!NPVARIANT_IS_STRING(stringObj))
        return NULL;

    NPString npString = NPVARIANT_TO_STRING(stringObj);
    char *retval = new char[npString.UTF8Length + 1];
    strncpy(retval, npString.UTF8Characters, npString.UTF8Length);
    retval[npString.UTF8Length] = 0;

    return retval;
}

static void getNPRuntimeObjects(NPP instance)
{
    NPObject *windowObj = NULL;
    if (browser.getvalue(instance, NPNVWindowNPObject, &windowObj) != NPERR_NO_ERROR)
        windowObj = NULL;

    NPObject *pluginElementObj = NULL;
    if (browser.getvalue(instance, NPNVPluginElementNPObject, &pluginElementObj) != NPERR_NO_ERROR)
        pluginElementObj = NULL;

    NPIdentifier toString_id = browser.getstringidentifier("toString");
    char *windowStr = NULL, *pluginElementStr = NULL;
    if (windowObj) {
        NPVariant windowStrVar;
        if (browser.invoke(instance, windowObj, toString_id, NULL, 0, &windowStrVar)) {
            windowStr = getUTF8String(windowStrVar);
            browser.releasevariantvalue(&windowStrVar);
        }
    }
    if (pluginElementObj) {
        NPVariant pluginElementStrVar;
        if (browser.invoke(instance, pluginElementObj, toString_id, NULL, 0, &pluginElementStrVar)) {
            pluginElementStr = getUTF8String(pluginElementStrVar);
            browser.releasevariantvalue(&pluginElementStrVar);
        }
    }

    char *docStr = NULL, *urlStr = NULL, *docURIStr = NULL;
    if (windowObj) {
        NPIdentifier doc_id = browser.getstringidentifier("document");
        NPVariant docVar;
        if (browser.getproperty(instance, windowObj, doc_id, &docVar)) {
            NPObject *docObj = NPVARIANT_TO_OBJECT(docVar);
            if (docObj) {
                NPVariant docStrVar;
                if (browser.invoke(instance, docObj, toString_id, NULL, 0, &docStrVar)) {
                    docStr = getUTF8String(docStrVar);
                    browser.releasevariantvalue(&docStrVar);
                }

                NPIdentifier url_id = browser.getstringidentifier("URL");
                NPVariant urlVar;
                if (browser.getproperty(instance, docObj, url_id, &urlVar)) {
                    urlStr = getUTF8String(urlVar);
                    browser.releasevariantvalue(&urlVar);
                }

                NPIdentifier docURI_id = browser.getstringidentifier("documentURI");
                NPVariant docURIVar;
                if (browser.getproperty(instance, docObj, docURI_id, &docURIVar)) {
                    docURIStr = getUTF8String(docURIVar);
                    browser.releasevariantvalue(&docURIVar);
                }
            }
            browser.releasevariantvalue(&docVar);
        }
    }

    NSLog(@"DebugEventsPlugin: getNPRuntimeObjects: Window %s, PluginElement %s, document %s, document.URL %s, document.documentURI %s",
          windowStr, pluginElementStr, docStr, urlStr, docURIStr);

    if (windowStr)
        delete windowStr;
    if (pluginElementStr)
        delete pluginElementStr;
    if (docStr)
        delete docStr;
    if (urlStr)
        delete urlStr;
    if (docURIStr)
        delete docURIStr;

    if (windowObj)
        browser.releaseobject(windowObj);
    if (pluginElementObj)
        browser.releaseobject(pluginElementObj);
}

NPError NPP_New(NPMIMEType pluginType, NPP instance, uint16_t mode, int16_t argc, char* argn[], char* argv[], NPSavedData* saved)
{
    NSLog(@"DebugEventsPlugin: NPP_New() start: pluginType %s, instance 0x%.8X, mode %s",
           pluginType, (unsigned) instance, modeString(mode));
    if (!instance) {
        NSLog(@"***NULL instance!***");
        return NPERR_INVALID_INSTANCE_ERROR;
    }

    // Create per-instance storage
    PluginObject *obj = (PluginObject *)malloc(sizeof(PluginObject));
    bzero(obj, sizeof(PluginObject));

    obj->npp = instance;
    instance->pdata = obj;

    // Ask the browser if it supports the Cocoa event model
    NPBool supportsCocoa;
    if (browser.getvalue(instance, NPNVsupportsCocoaBool, &supportsCocoa) != NPERR_NO_ERROR)
        supportsCocoa = FALSE;
    // Needed for Opera
    if ((supportsCocoa != TRUE) && (supportsCocoa != FALSE))
        supportsCocoa = FALSE;

    // If the browser does support the Cocoa event model, check if that
    // setting has been overridden.
    if (supportsCocoa) {
        NPBool hasOverride = FALSE;
        NPBool override = TRUE;
        checkForOverride(SUPPORT_COCOA, &override, &hasOverride, argc, argn, argv);
        if (hasOverride)
            supportsCocoa = override;
    }

    obj->supportsCocoa = supportsCocoa;

    // Enable the Cocoa event model if the browser supports it and
    // this hasn't been overridden.
    if (supportsCocoa)
        browser.setvalue(instance, NPPVpluginEventModel, (void *)NPEventModelCocoa);

    // Ask the browser if it supports the CoreGraphics drawing model
    NPBool supportsCoreGraphics;
    if (browser.getvalue(instance, NPNVsupportsCoreGraphicsBool, &supportsCoreGraphics) != NPERR_NO_ERROR)
        supportsCoreGraphics = FALSE;
    // Needed for Opera
    if ((supportsCoreGraphics != TRUE) && (supportsCoreGraphics != FALSE))
        supportsCoreGraphics = FALSE;

    // If the browser does support the CoreGraphics drawing model, check if
    // that setting has been overridden.  But don't do this if we've enabled
    // the Cocoa event model -- Safari (the only browser that currently
    // supports the Cocoa event model) currently crashes if we turn on the
    // Cocoa event model but not the CoreGraphics drawing model.
    if (supportsCoreGraphics && !supportsCocoa) {
        NPBool hasOverride = FALSE;
        NPBool override = TRUE;
        checkForOverride(SUPPORT_CORE_GRAHPICS, &override, &hasOverride, argc, argn, argv);
        if (hasOverride)
            supportsCoreGraphics = override;
    }

    obj->supportsCoreGraphics = supportsCoreGraphics;

    // Enable the CoreGraphics drawing model if the browser supports it and
    // this hasn't been overridden.
    if (supportsCoreGraphics)
        browser.setvalue(instance, NPPVpluginDrawingModel, (void *)NPDrawingModelCoreGraphics);

    obj->mode = mode;

    TestNPObject* scriptableObject = (TestNPObject*) browser.createobject(instance, &sSOClass);
    if (!scriptableObject) {
        NSLog(@"DebugEventsPlugin: NPP_New(): NPN_CreateObject failed!");
        free(obj);
        return NPERR_GENERIC_ERROR;
    }
    scriptableObject->npp = instance;
    obj->scriptableObject = scriptableObject;

    getNPRuntimeObjects(instance);

    NSLog(@"DebugEventsPlugin: NPP_New() finish: supportsCoreGraphics %i, supportsCocoa %i",
           supportsCoreGraphics, supportsCocoa);

    return NPERR_NO_ERROR;
}

static NPBool checkInstance(NPP instance)
{
    if (!instance) {
        NSLog(@"***NULL instance!***");
        return FALSE;
    } else if (!instance->pdata) {
        NSLog(@"***Instance not initialized!***");
        return FALSE;
    } else if (instance->pdata == (void *) -1L) {
        NSLog(@"***Instance destroyed!***");
        return FALSE;
    }
    return true;
}

NPError NPP_Destroy(NPP instance, NPSavedData** save)
{
    NSLog(@"DebugEventsPlugin: NPP_Destroy(): instance 0x%.8X", (unsigned) instance);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    // Free per-instance storage
    PluginObject *obj = (PluginObject *) instance->pdata;
    instance->pdata = (void *) -1L; // Mark destroyed

    browser.releaseobject(obj->scriptableObject);

    [obj->debugInfo release];

    free(obj);

    return NPERR_NO_ERROR;
}

NSWindow *cocoaWindowForCGContext(CGContextRef aContext)
{
    if (!aContext)
        return nil;

    int windowCount;
    NSCountWindows(&windowCount);
    int *windowList = (int *) malloc(sizeof(int) * windowCount);
    if (!windowList)
        return nil;
    NSWindowList(windowCount, windowList);

    for (int i = 0; i < windowCount; ++i) {
        NSWindow *aWindow = [NSApp windowWithWindowNumber:windowList[i]];
        if (!aWindow)
            continue;
        if (![aWindow respondsToSelector:@selector(graphicsContext)]) {
            free(windowList);
            return nil;
        }
        NSGraphicsContext *aNSContext = [aWindow graphicsContext];
        if (aNSContext && ([aNSContext graphicsPort] == aContext)) {
            free(windowList);
            return aWindow;
        }
    }

    free(windowList);
    return nil;
}

NSWindow *cocoaWindowForCarbonWindow(WindowRef aCarbonWindow)
{
    if (!aCarbonWindow)
        return nil;

    int windowCount;
    NSCountWindows(&windowCount);
    int *windowList = (int *) malloc(sizeof(int) * windowCount);
    if (!windowList)
        return nil;
    NSWindowList(windowCount, windowList);

    for (int i = 0; i < windowCount; ++i) {
        NSWindow *aWindow = [NSApp windowWithWindowNumber:windowList[i]];
        if ([NSWindow existingCarbonWindowFor:aWindow] == aCarbonWindow) {
            free(windowList);
            return aWindow;
        }

    }

    free(windowList);
    return nil;
}

void __attribute__((noinline)) setWindow_cgContextWrong_break()
{
}
void __attribute__((noinline)) setWindow_cgContextRight_break()
{
}

NPError NPP_SetWindow(NPP instance, NPWindow* window)
{
    NSLog(@"DebugEventsPlugin: NPP_SetWindow(): instance 0x%.8X, window 0x%.8X "
           "x %i y %i width %u height %u, clipRect (top %u left %u bottom %u right %u), type %s",
           (unsigned) instance, window ? (unsigned) window->window : 0,
           window ? window->x : 0, window ? window->y : 0,
           window ? window->width : 0, window ? window->height : 0,
           window ? window->clipRect.top : 0, window ? window->clipRect.left : 0,
           window ? window->clipRect.bottom : 0, window ? window->clipRect.right : 0,
           window ? typeString(window->type) : "notype");
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    PluginObject *obj = (PluginObject *) instance->pdata;
    obj->window = *window;

    if (obj->window.window) {
        if (obj->supportsCoreGraphics) {
            NP_CGContext *cgWindow = (NP_CGContext *) obj->window.window;
            NSLog(@"DebugEventsPlugin: NPP_SetWindow(): context 0x%.8X, window 0x%.8X",
                  cgWindow->context, cgWindow->window);
            NSWindow *cgContextWindow = cocoaWindowForCGContext(cgWindow->context);
            NSWindow *pluginWindow = cocoaWindowForCarbonWindow((WindowRef) cgWindow->window);
            NSView *drawingView = [NSView focusView];
            NSWindow *drawingWindow = [drawingView window];
            NSLog(@"DebugEventsPlugin: NPP_SetWindow(): cgContextWindow %@, pluginWindow %@, drawingWindow %@, drawingView %@",
                  cgContextWindow, pluginWindow, drawingWindow, drawingView);
            if (cgContextWindow && pluginWindow) {
                if (cgContextWindow == pluginWindow) {
                    setWindow_cgContextRight_break();
                } else {
                    setWindow_cgContextWrong_break();
                }
            }
        } else {
            NP_Port *qdWindow = (NP_Port *) obj->window.window;
            NSWindow *pluginWindow = cocoaWindowForCarbonWindow(::GetWindowFromPort(qdWindow->port));
            NSView *drawingView = [NSView focusView];
            NSWindow *drawingWindow = [drawingView window];
            NSLog(@"DebugEventsPlugin: NPP_SetWindow(): port 0x%.8X, portx %i, porty %i, pluginWindow %@, drawingWindow %@, drawingView %@",
                  qdWindow->port, qdWindow->portx, qdWindow->porty, pluginWindow, drawingWindow, drawingView);
        }
    } else {
        NSView *drawingView = [NSView focusView];
        NSWindow *drawingWindow = [drawingView window];
        NSLog(@"DebugEventsPlugin: NPP_SetWindow(): drawingWindow %@, drawingView %@", drawingWindow, drawingView);
    }

    return NPERR_NO_ERROR;
}
 

NPError NPP_NewStream(NPP instance, NPMIMEType type, NPStream* stream, NPBool seekable, uint16_t* stype)
{
    NSLog(@"DebugEventsPlugin: NPP_NewStream(): instance 0x%.8X, stream 0x%.8X, type %s",
           (unsigned) instance, (unsigned) stream, type);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    *stype = NP_ASFILEONLY;
    return NPERR_NO_ERROR;
}

NPError NPP_DestroyStream(NPP instance, NPStream* stream, NPReason reason)
{
    NSLog(@"DebugEventsPlugin: NPP_DestroyStream(): instance 0x%.8X, stream 0x%.8X",
           (unsigned) instance, (unsigned) stream);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    return NPERR_NO_ERROR;
}

int32_t NPP_WriteReady(NPP instance, NPStream* stream)
{
    NSLog(@"DebugEventsPlugin: NPP_WriteReady(): instance 0x%.8X, stream 0x%.8X",
           (unsigned) instance, (unsigned) stream);
    checkInstance(instance);

    return 0;
}

int32_t NPP_Write(NPP instance, NPStream* stream, int32_t offset, int32_t len, void* buffer)
{
    NSLog(@"DebugEventsPlugin: NPP_Write(): instance 0x%.8X, stream 0x%.8X, offset %i, len %i",
           (unsigned) instance, (unsigned) stream, offset, len);
    checkInstance(instance);

    return 0;
}

void NPP_StreamAsFile(NPP instance, NPStream* stream, const char* fname)
{
    NSLog(@"DebugEventsPlugin: NPP_StreamAsFile(): instance 0x%.8X, stream 0x%.8X, fname %s",
           (unsigned) instance, (unsigned) stream, fname);
    checkInstance(instance);
}

void NPP_Print(NPP instance, NPPrint* platformPrint)
{
    NSLog(@"DebugEventsPlugin: NPP_Print(): instance 0x%.8X, mode %s",
           (unsigned) instance, platformPrint ? modeString(platformPrint->mode) : "NULL");
    checkInstance(instance);
}

static NPBool startDrawQD(PluginObject *obj)
{
    if (!obj || obj->supportsCoreGraphics)
        return FALSE;

    NP_Port *npport = (NP_Port *) obj->window.window;
    CGrafPtr ourPort = npport->port;

    if (obj->window.clipRect.left < obj->window.clipRect.right) {
        ::GetPort(&obj->savedPortQD);
        ::SetPort((GrafPtr) ourPort);
        Rect portRect;
        ::GetPortBounds(ourPort, &portRect);
        obj->savedPortTopQD = portRect.top;
        obj->savedPortLeftQD = portRect.left;
        ::GetClip(obj->savedClipQD);

        Rect revealedRectQD;
        revealedRectQD.top = obj->window.clipRect.top + npport->porty;
        revealedRectQD.left = obj->window.clipRect.left + npport->portx;
        revealedRectQD.bottom = obj->window.clipRect.bottom + npport->porty;
        revealedRectQD.right = obj->window.clipRect.right + npport->portx;
        ::SetOrigin(npport->portx, npport->porty);
        ::ClipRect(&revealedRectQD);

        return TRUE;
    }

    return FALSE;
}

static void endDrawQD(PluginObject *obj)
{
    if (!obj)
        return;
    ::SetOrigin(obj->savedPortLeftQD, obj->savedPortTopQD);
    ::SetClip(obj->savedClipQD);
    ::SetPort(obj->savedPortQD);
}

static void drawStringQD(char* text,
                         Rect drawRect)
{
    if (!text)
        return;

    char *buffer = (char *) malloc(strlen(text) + 1);
    if (!buffer)
        return;
    strcpy(buffer, text);

    char *token = strtok(buffer, "\n");
    if (!token) {
        free(buffer);
        return;
    }

    ::TextFont(1);
    ::TextFace(bold);
    ::TextMode(srcCopy);
    ::TextSize(12);

    FontInfo fontInfo;
    ::GetFontInfo(&fontInfo);

    short textHeight = fontInfo.ascent + fontInfo.descent + fontInfo.leading;

    short drawPosX = drawRect.left + 5;
    short drawPosY = drawRect.top + textHeight;

    while (token) {
        short length = strlen(token);
        ::MoveTo(drawPosX, drawPosY);
        ::DrawText(token, 0, length);
        drawPosY += textHeight;
        token = strtok(NULL, "\n");
    }

    free(buffer);
}

static void doDrawQD(PluginObject *obj)
{
    Rect drawRect;
    RGBColor black = { 0x0000, 0x0000, 0x0000 };
    RGBColor white = { 0xFFFF, 0xFFFF, 0xFFFF };
    RGBColor lightGray = { 0xA000, 0xA000, 0xA000 };
    //RGBColor lightYellow = { 0xF000, 0xF000, 0xD000 };
    //RGBColor yellowGray = { 0xA000, 0xA000, 0x8000 };

    char *pTheText = (char *) [obj->debugInfo UTF8String];

    drawRect.top = 0;
    drawRect.left = 0;
    drawRect.bottom = obj->window.height;
    drawRect.right = obj->window.width;

    ::PenNormal();
    ::RGBForeColor(&black);
    if (obj->hasFocus) {
        ::RGBBackColor(&white);
    } else {
        ::RGBBackColor(&lightGray);
    }

    Pattern qdPattern;
    ::GetQDGlobalsWhite(&qdPattern);
    ::FillRect(&drawRect, &qdPattern);

    if (obj->hasFocus)
        ::FrameRect(&drawRect);
    drawStringQD(pTheText, drawRect);
}

static void handleDraw(PluginObject *obj, NPCocoaEvent *event = NULL)
{
    if (obj->supportsCoreGraphics) {
        NSGraphicsContext *context = NULL;
        if (event) {
            context = [NSGraphicsContext graphicsContextWithGraphicsPort:event->data.draw.context
                                                                 flipped:YES];
        } else {
            context = [NSGraphicsContext graphicsContextWithGraphicsPort:((NP_CGContext *)obj->window.window)->context
                                                                 flipped:YES];
		}
        if (!context) {
            NSLog(@"handleDraw(): ***NULL graphics context!***");
            return;
        }

        NSWindow *cgContextWindow = nil;
        if (event) {
            cgContextWindow = cocoaWindowForCGContext(event->data.draw.context);
        } else {
            cgContextWindow = cocoaWindowForCGContext(((NP_CGContext *)obj->window.window)->context);
        }
        NSView *drawingView = [NSView focusView];
        NSWindow *drawingWindow = [drawingView window];
        NSLog(@"DebugEventsPlugin: handleDraw(): context 0x%.8X, cgContextWindow %@, drawingWindow %@",
              [context graphicsPort], cgContextWindow, drawingWindow);
        NSLog(@"DebugEventsPlugin: handleDraw(): drawingView %@ and its superviews:",
              drawingView);
        NSView *superview = [drawingView superview];
        while (superview) {
            NSLog(@"  %@", superview);
            superview = [superview superview];
        }

        NSGraphicsContext *oldContext = [[NSGraphicsContext currentContext] retain];
        [NSGraphicsContext setCurrentContext:context];

        NSRect rect = NSMakeRect(0, 0, obj->window.width, obj->window.height);

        if (obj->hasFocus) {
            if (obj->supportsCocoa) {
                if (obj->windowHasFocus) {
                    [[NSColor whiteColor] set];
                } else {
                    // Light yellow
                    [[NSColor colorWithCalibratedRed:61440.0/65536.0
                                               green:61440.0/65536.0
                                                blue:53248.0/65536.0
                                               alpha:1.0] set];
                }
            } else {
                [[NSColor whiteColor] set];
            }
        } else {
            if (obj->supportsCocoa) {
                if (obj->windowHasFocus) {
                    [[NSColor lightGrayColor] set];
                } else {
                    // Yellow gray
                    [[NSColor colorWithCalibratedRed:40960.0/65536.0
                                               green:40960.0/65536.0
                                                blue:32768.0/65536.0
                                               alpha:1.0] set];
                }
            } else {
                [[NSColor lightGrayColor] set];
            }
        }
        [NSBezierPath fillRect:rect];

        // If the plugin has focus, draw a focus indicator
        if (obj->hasFocus) {
            [[NSColor blackColor] set];
            NSBezierPath *path = [NSBezierPath bezierPathWithRect:rect];
            [path setLineWidth:5];
            [path stroke];
        }

        [obj->debugInfo drawAtPoint:NSMakePoint(10, 10) withAttributes:nil];

        [NSGraphicsContext setCurrentContext:oldContext];
        [oldContext release];
    } else {
        if (startDrawQD(obj)) {
            doDrawQD(obj);
            endDrawQD(obj);
        }
    }
}

static NSString *eventTypeCocoa(NPCocoaEventType type)
{
    switch (type) {
        case NPCocoaEventScrollWheel:
            return @"NPCocoaEventScrollWheel";
        case NPCocoaEventMouseDown:
            return @"NPCocoaEventMouseDown";
        case NPCocoaEventMouseUp:
            return @"NPCocoaEventMouseUp";            
        case NPCocoaEventMouseMoved:
            return @"NPCocoaEventMouseMoved";            
        case NPCocoaEventMouseDragged:
            return @"NPCocoaEventMouseDragged";            
        case NPCocoaEventMouseEntered:
            return @"NPCocoaEventMouseEntered";            
        case NPCocoaEventMouseExited:
            return @"NPCocoaEventMouseExited";
        case NPCocoaEventKeyDown:
            return @"NPCocoaEventKeyDown";
        case NPCocoaEventKeyUp:
            return @"NPCocoaEventKeyUp";
        case NPCocoaEventFlagsChanged:
            return @"NPCocoaEventFlagsChanged";
        case NPCocoaEventFocusChanged:
            return @"NPCocoaEventFocusChanged";
        case NPCocoaEventWindowFocusChanged:
            return @"NPCocoaEventWindowFocusChanged";
        case NPCocoaEventTextInput:
            return @"NPCocoaEventTextInput";
        default:
            return @"Unknown";
    }    
}

static NSString *eventTypeQD(EventKind type)
{
    switch (type) {
        case mouseDown:
            return @"mouseDown";
        case mouseUp:
            return @"mouseUp";
        case osEvt:
            return @"mouseMoved";
        case keyDown:
            return @"keyDown";
        case keyUp:
            return @"keyUp";
        case NPEventType_GetFocusEvent:
            return @"NPEventType_GetFocusEvent";
        case NPEventType_LoseFocusEvent:
            return @"NPEventType_LoseFocusEvent";
        default:
            return @"Unknown";
    }    
}

static void invalidatePlugin(PluginObject *obj)
{
    NPRect rect;
    rect.left = 0;
    rect.top = 0;
    rect.right = obj->window.width;
    rect.bottom = obj->window.height;
    
    browser.invalidaterect(obj->npp, &rect);    
}

static void handleMouseEvent(PluginObject *obj, void *event)
{
    NSString *debugInfo = nil;
    if (obj->supportsCocoa) {
        NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8x\n"
                                               "Coordinates: (%g, %g)\n"
                                               "Button number: %d\n"
                                               "Click count: %d\n"
                                               "Delta: (%g, %g, %g)\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                               eventTypeCocoa(cocoaEvent->type),
                                               cocoaEvent->data.mouse.modifierFlags,
                                               cocoaEvent->data.mouse.pluginX,
                                                   cocoaEvent->data.mouse.pluginY,
                                               cocoaEvent->data.mouse.buttonNumber,
                                               cocoaEvent->data.mouse.clickCount,
                                               cocoaEvent->data.mouse.deltaX,
                                                   cocoaEvent->data.mouse.deltaY,
                                                   cocoaEvent->data.mouse.deltaZ,
                                               (cocoaEvent->data.mouse.modifierFlags &
                                                   NSAlternateKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.mouse.modifierFlags &
                                                   NSControlKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.mouse.modifierFlags &
                                                   NSCommandKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.mouse.modifierFlags &
                                                   NSShiftKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.mouse.modifierFlags &
                                                   NSAlphaShiftKeyMask) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type: %@"
                                ", Modifiers: 0x%.8x"
                                ", Coordinates: (%g, %g)"
                                ", Button number: %d"
                                ", Click count: %d"
                                ", Delta: (%g, %g, %g)",
                                (unsigned) obj->npp,
                                eventTypeCocoa(cocoaEvent->type),
                                cocoaEvent->data.mouse.modifierFlags,
                                cocoaEvent->data.mouse.pluginX,
                                    cocoaEvent->data.mouse.pluginY,
                                cocoaEvent->data.mouse.buttonNumber,
                                cocoaEvent->data.mouse.clickCount,
                                cocoaEvent->data.mouse.deltaX,
                                    cocoaEvent->data.mouse.deltaY,
                                    cocoaEvent->data.mouse.deltaZ);
    } else {
        EventRecord *qdEvent = (EventRecord *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8X\n"
                                               "Coordinates: (%d, %d)\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                               eventTypeQD(qdEvent->what),
                                               qdEvent->modifiers,
                                               qdEvent->where.h, qdEvent->where.v,
                                               ((qdEvent->modifiers & optionKey) ||
                                                   (qdEvent->modifiers & rightOptionKey))
                                                   ? "Down" : "Up",
                                               ((qdEvent->modifiers & controlKey) ||
                                                   (qdEvent->modifiers & rightControlKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & cmdKey) ? "Down" : "Up",
                                               ((qdEvent->modifiers & shiftKey) ||
                                                   (qdEvent->modifiers & rightShiftKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & alphaLock) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type: %@"
                                ", Modifiers: 0x%.8X"
                                ", Coordinates: (%d, %d)",
                                (unsigned) obj->npp,
                                eventTypeQD(qdEvent->what),
                                qdEvent->modifiers,
                                qdEvent->where.h, qdEvent->where.v);
    }
    [obj->debugInfo release];
    obj->debugInfo = [debugInfo retain];
 
    invalidatePlugin(obj);
}

static BOOL lineChanger(unichar charCode)
{
    BOOL retval = NO;
    switch (charCode) {
        case 10:
        case 13:
            retval = YES;
            break;
        default:
            break;
	}
    return retval;
}

static void handleKeyboardEvent(PluginObject *obj, void *event)
{
    NSString *debugInfo = nil;
    if (obj->supportsCocoa) {
        NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;
        // Work around FF bug that can leave these fields unitialized.
        if (cocoaEvent->type == NPCocoaEventFlagsChanged) {
            cocoaEvent->data.key.characters = NULL;
            cocoaEvent->data.key.charactersIgnoringModifiers = NULL;
        }
        unichar charCode = [(NSString *)cocoaEvent->data.key.characters characterAtIndex:0];
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8x\n"
                                               "Characters: %@\n"
                                               "Character code: %d\n"
                                               "Characters ignoring modifiers: %@\n"
                                               "Is a repeat: %@\n"
                                               "Key code: %d\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                                eventTypeCocoa(cocoaEvent->type),
                                                cocoaEvent->data.key.modifierFlags,
                                                cocoaEvent->data.key.characters,
                                                charCode,
                                                cocoaEvent->data.key.charactersIgnoringModifiers,
                                                cocoaEvent->data.key.isARepeat ? @"YES" : @"NO",
                                                cocoaEvent->data.key.keyCode,
                                               (cocoaEvent->data.key.modifierFlags &
                                                   NSAlternateKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.key.modifierFlags &
                                                   NSControlKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.key.modifierFlags &
                                                   NSCommandKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.key.modifierFlags &
                                                   NSShiftKeyMask) ? "Down" : "Up",
                                               (cocoaEvent->data.key.modifierFlags &
                                                   NSAlphaShiftKeyMask) ? "Down" : "Up"];
        if (!lineChanger(charCode)) {
            NSLog(@"DebugEventsPlugin (0x%.8X): Type %@"
                                    ", Modifiers: 0x%.8x"
                                    ", Characters: %@"
                                    ", Character code: %d"
                                    ", Characters ignoring modifiers: %@"
                                    ", Is a repeat: %@"
                                    ", Key code: %d",
                                    (unsigned) obj->npp,
                                    eventTypeCocoa(cocoaEvent->type),
                                    cocoaEvent->data.key.modifierFlags,
                                    cocoaEvent->data.key.characters,
                                    charCode,
                                    cocoaEvent->data.key.charactersIgnoringModifiers,
                                    cocoaEvent->data.key.isARepeat ? @"YES" : @"NO",
                                    cocoaEvent->data.key.keyCode);
        } else {
            NSLog(@"DebugEventsPlugin (0x%.8X): Type %@"
                                    ", Modifiers: 0x%.8x"
                                    ", Character code: %d"
                                    ", Is a repeat: %@"
                                    ", Key code: %d",
                                    (unsigned) obj->npp,
                                    eventTypeCocoa(cocoaEvent->type),
                                    cocoaEvent->data.key.modifierFlags,
                                    charCode,
                                    cocoaEvent->data.key.isARepeat ? @"YES" : @"NO",
                                    cocoaEvent->data.key.keyCode);
        }
    } else {
        EventRecord *qdEvent = (EventRecord *) event;
        debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                               "Modifiers: 0x%.8X\n"
                                               "Character code: %d\n"
                                               "Key code: %d\n"
                                               "Option key: %s\n"
                                               "Control key: %s\n"
                                               "Command key: %s\n"
                                               "Shift key: %s\n"
                                               "Caps lock: %s",
                                               eventTypeQD(qdEvent->what),
                                               qdEvent->modifiers,
                                               qdEvent->message & charCodeMask,
                                               (qdEvent->message & keyCodeMask) >> 8,
                                               ((qdEvent->modifiers & optionKey) ||
                                                   (qdEvent->modifiers & rightOptionKey))
                                                   ? "Down" : "Up",
                                               ((qdEvent->modifiers & controlKey) ||
                                                   (qdEvent->modifiers & rightControlKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & cmdKey) ? "Down" : "Up",
                                               ((qdEvent->modifiers & shiftKey) ||
                                                   (qdEvent->modifiers & rightShiftKey))
                                                   ? "Down" : "Up",
                                               (qdEvent->modifiers & alphaLock) ? "Down" : "Up"];
        NSLog(@"DebugEventsPlugin (0x%.8X): Type %@"
                                ", Modifiers: 0x%.8X"
                                ", Character code: %d"
                                ", Key code: %d",
                                (unsigned) obj->npp,
                                eventTypeQD(qdEvent->what),
                                qdEvent->modifiers,
                                qdEvent->message & charCodeMask,
                                (qdEvent->message & keyCodeMask) >> 8);
    }

    [obj->debugInfo release];
    obj->debugInfo = [debugInfo retain];

    invalidatePlugin(obj);
}

static void handleTextEvent(PluginObject *obj, void *event)
{
    if (!obj->supportsCocoa)
        return;

    NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;

    NSString *debugInfo = [NSString stringWithFormat:@"Type: %@\n"
                                                     "Text: %@",
                                                     eventTypeCocoa(cocoaEvent->type),
                                                     cocoaEvent->data.text.text];

    NSLog(@"DebugEventsPlugin (0x%.8X): Type %@"
                            ", Text: %@",
                            (unsigned) obj->npp,
                            eventTypeCocoa(cocoaEvent->type),
                            cocoaEvent->data.text.text);

    [obj->debugInfo release];
    obj->debugInfo = [debugInfo retain];

    invalidatePlugin(obj);
}

static void handleGetFocusEvent(PluginObject *obj)
{
    if (!obj)
        return;

    if (obj->hasFocus) {
        NSLog(@"DebugEventsPlugin (0x%.8X): Redundant getFocus event", (unsigned) obj->npp);
    } else {
        NSLog(@"DebugEventsPlugin (0x%.8X): getFocus", (unsigned) obj->npp);
        obj->hasFocus = TRUE;
        invalidatePlugin(obj);
    }
}

static void handleLoseFocusEvent(PluginObject *obj)
{
    if (!obj)
        return;

    if (!obj->hasFocus) {
        NSLog(@"DebugEventsPlugin (0x%.8X): Redundant loseFocus event", (unsigned) obj->npp);
    } else {
        NSLog(@"DebugEventsPlugin (0x%.8X): loseFocus", (unsigned) obj->npp);
        obj->hasFocus = FALSE;
        invalidatePlugin(obj);
    }
}

static void handleWindowGetFocusEvent(PluginObject *obj)
{
    if (!obj)
        return;

    if (obj->windowHasFocus) {
        NSLog(@"DebugEventsPlugin (0x%.8X): Redundant windowGetFocus event", (unsigned) obj->npp);
    } else {
        NSLog(@"DebugEventsPlugin (0x%.8X): windowGetFocus", (unsigned) obj->npp);
        obj->windowHasFocus = TRUE;
        invalidatePlugin(obj);
    }
}

static void handleWindowLoseFocusEvent(PluginObject *obj)
{
    if (!obj)
        return;

    if (!obj->windowHasFocus) {
        NSLog(@"DebugEventsPlugin (0x%.8X): Redundant windowLoseFocus event", (unsigned) obj->npp);
    } else {
        NSLog(@"DebugEventsPlugin (0x%.8X): windowLoseFocus", (unsigned) obj->npp);
        obj->windowHasFocus = FALSE;
        invalidatePlugin(obj);
    }
}

int16_t NPP_HandleEvent(NPP instance, void* event)
{
    if (!checkInstance(instance)) {
        NSLog(@"DebugEventsPlugin: NPP_HandleEvent() error return");
        return kNPEventNotHandled;
    }
    if (!event) {
        NSLog(@"DebugEventsPlugin: NPP_HandleEvent(): NULL 'event'!");
        return kNPEventNotHandled;
    }

    PluginObject *obj = (PluginObject *) instance->pdata;

    if (obj->supportsCocoa) {
        NPCocoaEvent *cocoaEvent = (NPCocoaEvent *) event;
        switch(cocoaEvent->type) {
            case NPCocoaEventFocusChanged:
                if (cocoaEvent->data.focus.hasFocus) {
                    handleGetFocusEvent(obj);
                } else {
                    handleLoseFocusEvent(obj);
                }
                return kNPEventHandled;

            case NPCocoaEventWindowFocusChanged:
                if (cocoaEvent->data.focus.hasFocus) {
                    handleWindowGetFocusEvent(obj);
                } else {
                    handleWindowLoseFocusEvent(obj);
                }
                return kNPEventHandled;

            case NPCocoaEventDrawRect:
                NSLog(@"DebugEventsPlugin: NPCocoaEventDrawRect: instance 0x%.8X, window 0x%.8X, dirtyRect (x %f y %f width %f height %f)",
                      (unsigned) instance, (unsigned) obj->window.window,
                      cocoaEvent->data.draw.x, cocoaEvent->data.draw.y,
                      cocoaEvent->data.draw.width, cocoaEvent->data.draw.height);
                handleDraw(obj, cocoaEvent);
                return kNPEventHandled;

            case NPCocoaEventKeyDown:
                handleKeyboardEvent(obj, event);
                return kNPEventStartIME;
            case NPCocoaEventKeyUp:
            case NPCocoaEventFlagsChanged:
                handleKeyboardEvent(obj, event);
                return kNPEventHandled;

            case NPCocoaEventTextInput:
                handleTextEvent(obj, event);
                return kNPEventHandled;

#ifdef LOG_MOUSE_MOVED_EVENTS
            case NPCocoaEventMouseMoved:
#endif
            case NPCocoaEventMouseDown:
            case NPCocoaEventMouseUp:
            case NPCocoaEventMouseEntered:
            case NPCocoaEventMouseExited:
            case NPCocoaEventMouseDragged:
            case NPCocoaEventScrollWheel:
                handleMouseEvent(obj, event);
                return kNPEventHandled;
            default:
                break;
        }
    } else {
        EventRecord *qdEvent = (EventRecord *) event;
        switch (qdEvent->what) {
#ifdef LOG_MOUSE_MOVED_EVENTS
            case osEvt: // Assume mouse-moved event (not 'suspend' or 'resume')
#endif
            case mouseDown:
            case mouseUp:
                handleMouseEvent(obj, event);
                return kNPEventHandled;
            case keyDown:
            case keyUp:
                handleKeyboardEvent(obj, event);
                return kNPEventHandled;
            case updateEvt:
                NSLog(@"DebugEventsPlugin: updateEvt: instance 0x%.8X, window 0x%.8X",
                      (unsigned) instance, (unsigned) obj->window.window);
                handleDraw(obj);
                return kNPEventHandled;
            case NPEventType_GetFocusEvent:
                handleGetFocusEvent(obj);
                return kNPEventHandled;
            case NPEventType_LoseFocusEvent:
                handleLoseFocusEvent(obj);
                return kNPEventHandled;
            default:
                break;
        }
    }
    
    return kNPEventNotHandled;
}

void NPP_URLNotify(NPP instance, const char* url, NPReason reason, void* notifyData)
{
    NSLog(@"DebugEventsPlugin: NPP_URLNotify(): instance 0x%.8X, url %s",
           (unsigned) instance, url);
    checkInstance(instance);
}

NPError NPP_GetValue(NPP instance, NPPVariable variable, void *value)
{
    NSLog(@"DebugEventsPlugin: NPP_GetValue(): instance 0x%.8X, variable %i",
           (unsigned) instance, variable);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;

    PluginObject *obj = (PluginObject *) instance->pdata;

    if (variable == NPPVpluginScriptableNPObject) {
        NPObject* object = obj->scriptableObject;
        browser.retainobject(object);
        *((NPObject**)value) = object;
        return NPERR_NO_ERROR;
    }

    return NPERR_GENERIC_ERROR;
}

NPError NPP_SetValue(NPP instance, NPNVariable variable, void *value)
{
    NSLog(@"DebugEventsPlugin: NPP_SetValue(): instance 0x%.8X, variable %i",
           (unsigned) instance, variable);
    if (!checkInstance(instance))
        return NPERR_INVALID_INSTANCE_ERROR;
    return NPERR_GENERIC_ERROR;
}

// Scriptable npruntime object methods

NPObject* scriptableAllocate(NPP npp, NPClass* aClass)
{
    TestNPObject* object = (TestNPObject*) browser.memalloc(sizeof(TestNPObject));
    if (!object)
        return NULL;
    memset(object, 0, sizeof(TestNPObject));
    return object;
}

void scriptableDeallocate(NPObject* npobj)
{
    browser.memfree(npobj);
}

void scriptableInvalidate(NPObject* npobj)
{
}

bool scriptableHasMethod(NPObject* npobj, NPIdentifier name)
{
    return false;
}

bool scriptableInvoke(NPObject* npobj, NPIdentifier name, const NPVariant* args, uint32_t argCount, NPVariant* result)
{
    return false;
}

bool scriptableInvokeDefault(NPObject* npobj, const NPVariant* args, uint32_t argCount, NPVariant* result)
{
    return false;
}

bool scriptableHasProperty(NPObject* npobj, NPIdentifier name)
{
    return false;
}

bool scriptableGetProperty(NPObject* npobj, NPIdentifier name, NPVariant* result)
{
    return false;
}

bool scriptableSetProperty(NPObject* npobj, NPIdentifier name, const NPVariant* value)
{
    return false;
}

bool scriptableRemoveProperty(NPObject* npobj, NPIdentifier name)
{
    return false;
}

bool scriptableEnumerate(NPObject* npobj, NPIdentifier** identifier, uint32_t* count)
{
    return false;
}

bool scriptableConstruct(NPObject* npobj, const NPVariant* args, uint32_t argCount, NPVariant* result)
{
    return false;
}
